@@ -18,7 +18,7 @@ import unittest
 import torch
 
 from torchimize.functions import lsq_gna_parallel, lsq_gna_parallel_plain
-from tests.emg_mm import *
+from tests.emg import *
 
 class ParallelOptimizationTest(unittest.TestCase):
 
@@ -32,7 +32,7 @@ class ParallelOptimizationTest(unittest.TestCase):
 
         # alpha, mu, sigma, eta
         gt_params_list = [[10, -1, 80, 5], [5, -2, 60, 5], [8, 0, 90, 3], [10, -1, 80, 5], [5, -2, 60, 5], [8, 0, 90, 3]]
-        initials_list = [[7.5, -.75, 10, 3], [6, -1, 50, 4], [7, -1, 80, 4], [7.5, -.75, 10, 3], [6, -1, 50, 4], [7, -1, 80, 4]]
+        initials_list = [[7.5, -.75, 40, 3], [6, -1, 50, 4], [7, -1, 80, 4], [7.5, -.75, 40, 3], [6, -1, 50, 4], [7, -1, 80, 4]]
         channel_num = len(gt_params_list)
         self.gt_params = torch.tensor(gt_params_list, dtype=torch.float64, device=self.device, requires_grad=False)
         self.initials = torch.tensor(initials_list, dtype=torch.float64, device=self.device, requires_grad=False)
@@ -119,7 +119,7 @@ class ParallelOptimizationTest(unittest.TestCase):
 
     def test_gna_emg_conditions(self):
 
-        for p in [self.batch_initials.float(), self.batch_initials.double()]:
+        for p in [self.batch_initials.clone().float(), self.batch_initials.clone().double()]:
 
             coeffs = lsq_gna_parallel(
                 p = p,
@@ -127,49 +127,53 @@ class ParallelOptimizationTest(unittest.TestCase):
                 jac_function = self.multi_jaco_batch,
                 args = (self.t.to(dtype=p.dtype), self.batch_data_channels.to(dtype=p.dtype)),
                 wvec = torch.ones(2, device=self.device, dtype=p.dtype, requires_grad=False),
-                l = .1,
-                gtol = 1e-6,
-                max_iter = 199,
+                l = 1,
+                ftol = 1e-8,
+                ptol = 1e-8,
+                gtol = 1e-8,
+                max_iter = 99,
             )
 
             # assertion
+            self.iter_gna_cond = len(coeffs)
+            self.assertTrue(self.iter_gna_cond < 100, 'Number of iterations exceeded 100')
             ret_params = torch.allclose(coeffs[-1].double(), self.gt_params, atol=1e-1)
             self.assertTrue(ret_params, 'Coefficients deviate')
-            eps = torch.sum(self.cost_batch(coeffs[-1].double(), t=self.t, y=self.batch_data_channels))
-            self.assertTrue(eps.cpu()/len(self.gt_params) < 1, 'Error exceeded 1')
-            self.assertTrue(len(coeffs) < 200, 'Number of iterations exceeded 200')
+            eps = torch.sum(self.cost_batch(coeffs[-1].double(), t=self.t, y=self.batch_data_channels)).cpu()
+            self.eps_gna_cond = torch.round(eps, decimals=4)
+            self.assertTrue(self.eps_gna_cond/len(self.gt_params) < 1, 'Error exceeded 1')
 
     def test_gna_emg_plain(self):
 
-
-        for p in [self.batch_initials.float(), self.batch_initials.double()]:
+        for p in [self.batch_initials.clone().float(), self.batch_initials.clone().double()]:
 
             multi_cost_batch_args = lambda p: self.multi_cost_batch(p_batch=p, t=self.t.to(dtype=p.dtype), y=self.batch_data_channels.to(dtype=p.dtype))
             multi_jaco_batch_args = lambda p: self.multi_jaco_batch(p_batch=p, t=self.t.to(dtype=p.dtype), data=self.batch_data_channels.to(dtype=p.dtype))
-            
+            self.iter_gna = 199
+
             coeffs = lsq_gna_parallel_plain(
                 p = p,
                 function = multi_cost_batch_args,
                 jac_function = multi_jaco_batch_args,
                 wvec = torch.ones(2, device=self.device, dtype=p.dtype, requires_grad=False),
-                l = .1,
-                max_iter = 199,
+                l = 1,
+                max_iter = self.iter_gna,
             )
 
             # assertion
             ret_params = torch.allclose(coeffs.double(), self.gt_params, atol=1e-1)
             self.assertTrue(ret_params, 'Coefficients deviate')
-            eps = torch.sum(multi_cost_batch_args(coeffs.double()))
-            self.assertTrue(eps.cpu()/len(self.gt_params) < 1, 'Error exceeded 1')
-            self.assertTrue(len(coeffs) < 200, 'Number of iterations exceeded 200')
+            eps = torch.sum(multi_cost_batch_args(coeffs.double())).cpu()
+            self.eps_gna = torch.round(eps, decimals=4)
+            self.assertTrue(self.eps_gna/len(self.gt_params) < 1, 'Error exceeded 1')
 
     def test_lma_emg_conditions(self):
 
-        from torchimize.functions.lma_fun_parallel import lsq_lma_parallel
+        from torchimize.functions.parallel.lma_fun_parallel import lsq_lma_parallel
 
-        for p in [self.batch_initials.float(), self.batch_initials.double()]:
+        for p in [self.batch_initials.clone().float(), self.batch_initials.clone().double()]:
             
-            for m in ['lev', 'marq']:
+            for m in ['lev']:
 
                 coeffs = lsq_lma_parallel(
                     p = p,
@@ -178,21 +182,70 @@ class ParallelOptimizationTest(unittest.TestCase):
                     args = (self.t.to(dtype=p.dtype), self.batch_data_channels.to(dtype=p.dtype)),
                     wvec = torch.ones(2, device=self.device, dtype=p.dtype, requires_grad=False),
                     meth = m,
-                    max_iter = 199,
+                    ftol = 1e-8,
+                    ptol = 1e-8,
+                    gtol = 1e-8,
+                    max_iter = 99,
                 )
 
                 # assertion
+                self.iter_lma_cond = len(coeffs)
+                self.assertTrue(self.iter_lma_cond < 100, 'Number of iterations exceeded 100')
                 ret_params = torch.allclose(coeffs[-1].double(), self.gt_params, atol=1e-1)
                 self.assertTrue(ret_params, 'Coefficients deviate')
-                eps = torch.sum(self.cost_batch(coeffs[-1].double(), t=self.t, y=self.batch_data_channels))
-                self.assertTrue(eps.cpu()/len(self.gt_params) < 1, 'Error exceeded 1')
-                self.assertTrue(len(coeffs) < 40, 'Number of iterations exceeded 40')
+                eps = torch.sum(self.cost_batch(coeffs[-1].double(), t=self.t, y=self.batch_data_channels)).cpu()
+                self.eps_lma_cond = torch.round(eps, decimals=4)
+                self.assertTrue(self.eps_lma_cond/len(self.gt_params) < 1, 'Error exceeded 1')
+
+    def scipy_fit(self):
+
+        from scipy.optimize import least_squares
+        
+        i = 0
+        p = self.batch_initials[i, ...].clone().float().cpu()
+        t = self.t.clone().cpu()
+        data = self.batch_data_channels[i, ...].to(dtype=p.dtype).cpu()
+
+        emg_model_args = lambda p: self.emg_model(p, t=t)
+        emg_jacob_args = lambda p: self.emg_jac(p, t=t, data=data)
+
+        coeffs = least_squares(
+                fun = emg_model_args,
+                x0 = p,
+                jac = emg_jacob_args,
+                method = 'lm',
+            ).x
+
+        return coeffs
+
+    def test_fun_dims(self):
+
+        from torchimize.functions import test_fun_dims_parallel
+
+        for p in [self.batch_initials.float(), self.batch_initials.double()]:
+
+            ret = test_fun_dims_parallel(
+                p = p,
+                function = self.multi_cost_batch,
+                jac_function = self.multi_jaco_batch,
+                args = (self.t.to(dtype=p.dtype), self.batch_data_channels.to(dtype=p.dtype)),
+                wvec = torch.ones(2, device=self.device, dtype=p.dtype, requires_grad=False),
+            )
+
+            self.assertTrue(ret, 'torch tensor dimensionality test failed')
 
     def test_all(self):
 
-        self.test_lma_emg_conditions()
-        self.test_gna_emg_conditions()
+        self.test_fun_dims()
         self.test_gna_emg_plain()
+        self.test_gna_emg_conditions()
+        self.test_lma_emg_conditions()
+
+        # assert LMA performance against GNA
+        self.assertTrue(self.eps_lma_cond <= self.eps_gna, 'LMA error is larger than that of GNA')
+        self.assertTrue(self.eps_lma_cond <= self.eps_gna_cond, 'LMA error is larger than that of GNA')
+        self.assertTrue(self.iter_lma_cond <= self.iter_gna, 'LMA required more iterations than GNA')
+        self.assertTrue(self.iter_lma_cond <= self.iter_gna_cond, 'LMA required more iterations than GNA')
 
 
 if __name__ == '__main__':
